# Copyright 2020 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Intent service wrapping padatious."""
import re
import string
from collections import defaultdict
from functools import lru_cache
from os.path import expanduser, isfile
from threading import Event, RLock
from typing import Optional, Dict, List, Union, Type

import snowballstemmer
from langcodes import closest_match
from ovos_config.config import Configuration
from ovos_config.meta import get_xdg_base

from ovos_bus_client.client import MessageBusClient
from ovos_bus_client.message import Message
from ovos_bus_client.session import SessionManager, Session
from ovos_padatious import IntentContainer
from ovos_padatious.domain_container import DomainIntentContainer
from ovos_padatious.match_data import MatchData as PadatiousIntent
from ovos_plugin_manager.templates.pipeline import ConfidenceMatcherPipeline, IntentHandlerMatch, IntentMatch
from ovos_utils import flatten_list
from ovos_utils.bracket_expansion import expand_template
from ovos_utils.fakebus import FakeBus
from ovos_utils.lang import standardize_lang_tag
from ovos_utils.list_utils import deduplicate_list
from ovos_utils.log import LOG, deprecated, log_deprecation
from ovos_utils.text_utils import remove_accents_and_punct
from ovos_utils.xdg_utils import xdg_data_home
import faulthandler

PadatiousIntentContainer = IntentContainer  # backwards compat

# for easy typing
PadatiousEngine = Union[Type[IntentContainer], Type[DomainIntentContainer]]



def normalize_utterances(utterances: List[str], lang: str, cast_to_ascii: bool = True,
                         keep_order: bool = True, stemmer: Optional['Stemmer'] = None) -> List[str]:
    """
    Normalize a list of utterances by collapsing whitespaces, removing accents and punctuation,
    and optionally stemming and deduplicating.

    Args:
        utterances (List[str]): The list of utterances to normalize.
        lang (str): The language code for stemming support.
        cast_to_ascii (bool): Whether to remove accented characters and punctuation. Default is True.
        keep_order (bool): Whether to preserve the order of utterances. Default is True.
        stemmer (Optional[Stemmer]): A stemmer object to stem the utterances (default is None).

    Returns:
        List[str]: The normalized list of utterances.
    """
    # Flatten the list if it's in old style tuple format
    utterances = flatten_list(utterances)  # Assuming flatten_list is defined elsewhere
    # Collapse multiple whitespaces into a single space
    utterances = [re.sub(r'\s+', ' ', u) for u in utterances]
    # Replace accented characters and punctuation if needed
    if cast_to_ascii:
        utterances = [remove_accents_and_punct(u) for u in utterances]
    # strip punctuation marks, that just causes duplicate training data
    utterances = [u.rstrip(string.punctuation) for u in utterances]
    # Stem words if stemmer is provided
    if stemmer is not None:
        utterances = stemmer.stem_sentences(utterances)
    # Deduplicate the list
    utterances = deduplicate_list(utterances, keep_order=keep_order)
    return utterances


class Stemmer:
    """
    A simple wrapper around the Snowball stemmer for various languages.

    Attributes:
        LANGS (dict): A dictionary mapping language codes to Snowball stemmer language names.
    """
    LANGS = {'ar': 'arabic', 'eu': 'basque', 'ca': 'catalan', 'da': 'danish', 'nl': 'dutch', 'en': 'english',
             'fi': 'finnish', 'fr': 'french', 'de': 'german', 'el': 'greek', 'hi': 'hindi', 'hu': 'hungarian',
             'id': 'indonesian', 'ga': 'irish', 'it': 'italian', 'lt': 'lithuanian', 'ne': 'nepali',
             'no': 'norwegian', 'pt': 'portuguese', 'ro': 'romanian', 'ru': 'russian', 'sr': 'serbian',
             'es': 'spanish', 'sv': 'swedish', 'ta': 'tamil', 'tr': 'turkish'}

    def __init__(self, lang: str):
        """
        Initialize the stemmer for a given language.

        Args:
            lang (str): The language code for stemming.

        Raises:
            ValueError: If the language is unsupported.
        """
        lang2 = closest_match(lang, list(self.LANGS))[0]
        if lang2 == "und":
            raise ValueError(f"unsupported language: {lang}")
        self.snowball = snowballstemmer.stemmer(self.LANGS[lang2])

    @classmethod
    def supports_lang(cls, lang: str) -> bool:
        """
        Check if the given language is supported by the stemmer.

        Args:
            lang (str): The language code to check.

        Returns:
            bool: True if the language is supported, False otherwise.
        """
        lang2 = closest_match(lang, list(cls.LANGS))[0]
        return lang2 != "und"

    def stem_sentence(self, sentence: str) -> str:
        """
        Stem a single sentence.

        Args:
            sentence (str): The sentence to stem.

        Returns:
            str: The stemmed sentence.
        """
        return _cached_stem_sentence(self.snowball, sentence)

    def stem_sentences(self, sentences: List[str]) -> List[str]:
        """
        Stem a list of sentences.

        Args:
            sentences (List[str]): The list of sentences to stem.

        Returns:
            List[str]: The list of stemmed sentences.
        """
        return [self.stem_sentence(s) for s in sentences]


@lru_cache()
def _cached_stem_sentence(stemmer, sentence: str) -> str:
    """
    Cache the stemming of a single sentence to optimize repeated calls.

    Args:
        stemmer: The stemmer instance to use.
        sentence (str): The sentence to stem.

    Returns:
        str: The stemmed sentence.
    """
    stems = stemmer.stemWords(sentence.split())
    return " ".join(stems)


class PadatiousMatcher:
    """Matcher class to avoid redundancy in padatious intent matching."""

    @deprecated("PadatiousMatcher class is deprecated!", "2.0.0")
    def __init__(self, service: 'PadatiousPipeline'):
        self.service = service

    def _match_level(self, utterances, limit, lang=None, message: Optional[Message] = None) -> Optional[IntentMatch]:
        """Match intent and make sure a certain level of confidence is reached.

        Args:
            utterances (list of tuples): Utterances to parse, originals paired
                                         with optional normalized version.
            limit (float): required confidence level.
        """
        m: IntentHandlerMatch = self.service._match_level(utterances, limit, lang, message)
        if not m:
            return None
        return IntentMatch("Padatious", m.match_type, m.match_data, m.skill_id, m.utterance)

    def match_high(self, utterances, lang=None, message=None) -> Optional[IntentMatch]:
        """Intent matcher for high confidence.

        Args:
            utterances (list of tuples): Utterances to parse, originals paired
                                         with optional normalized version.
        """
        return self._match_level(utterances, self.service.conf_high, lang, message)

    def match_medium(self, utterances, lang=None, message=None) -> Optional[IntentMatch]:
        """Intent matcher for medium confidence.

        Args:
            utterances (list of tuples): Utterances to parse, originals paired
                                         with optional normalized version.
        """
        return self._match_level(utterances, self.service.conf_med, lang, message)

    def match_low(self, utterances, lang=None, message=None) -> Optional[IntentMatch]:
        """Intent matcher for low confidence.

        Args:
            utterances (list of tuples): Utterances to parse, originals paired
                                         with optional normalized version.
        """
        return self._match_level(utterances, self.service.conf_low, lang, message)


class PadatiousPipeline(ConfidenceMatcherPipeline):
    """Service class for padatious intent matching."""

    def __init__(self, bus: Optional[Union[MessageBusClient, FakeBus]] = None,
                 config: Optional[Dict] = None,
                 engine_class: Optional[PadatiousEngine] = None):

        super().__init__(bus, config)
        try:
            faulthandler.enable()  # Enables crash logging
        except Exception:
            pass # happens in unittests and such
        self.lock = RLock()
        core_config = Configuration()
        self.lang = standardize_lang_tag(core_config.get("lang", "en-US"))
        langs = core_config.get('secondary_langs') or []
        langs = [standardize_lang_tag(l) for l in langs]
        if self.lang not in langs:
            langs.append(self.lang)

        self.conf_high = self.config.get("conf_high") or 0.95
        self.conf_med = self.config.get("conf_med") or 0.8
        self.conf_low = self.config.get("conf_low") or 0.5

        engine_class = engine_class or DomainIntentContainer if self.config.get("domain_engine") else IntentContainer
        LOG.info(f"Padatious class: {engine_class.__name__}")

        self.remove_punct = self.config.get("cast_to_ascii", False)
        use_stemmer = self.config.get("stem", False)
        self.engine_class = engine_class or IntentContainer
        intent_cache = expanduser(self.config.get('intent_cache') or
                                  f"{xdg_data_home()}/{get_xdg_base()}/intent_cache")
        if self.engine_class == DomainIntentContainer:
            # allow user to switch back and forth without retraining
            # cache is cheap, training isn't
            intent_cache += "_domain"
        if use_stemmer:
            intent_cache += "_stemmer"
        if self.remove_punct:
            intent_cache += "_normalized"
        self.containers = {lang: self.engine_class(cache_dir=f"{intent_cache}/{lang}",
                                                   disable_padaos=self.config.get("disable_padaos", False))
                           for lang in langs}

        # pre-load any cached intents
        for container in self.containers.values():
            try:
                container.instantiate_from_disk()
            except Exception as e:
                LOG.error(f"Failed to pre-load cached intents: {str(e)}")

        if use_stemmer:
            self.stemmers = {lang: Stemmer(lang)
                             for lang in langs if Stemmer.supports_lang(lang)}
        else:
            self.stemmers = {}

        self.first_train = Event()
        self.finished_training_event = Event()
        self.finished_training_event.set()  # is cleared when training starts

        self.registered_intents = []
        self.registered_entities = []
        self._skill2intent = defaultdict(list)
        self.max_words = 50  # if an utterance contains more words than this, don't attempt to match

        self.bus.on('padatious:register_intent', self.register_intent)
        self.bus.on('padatious:register_entity', self.register_entity)
        self.bus.on('detach_intent', self.handle_detach_intent)
        self.bus.on('detach_skill', self.handle_detach_skill)
        self.bus.on('intent.service.padatious.get', self.handle_get_padatious)
        self.bus.on('intent.service.padatious.manifest.get', self.handle_padatious_manifest)
        self.bus.on('intent.service.padatious.entities.manifest.get', self.handle_entity_manifest)
        self.bus.on('mycroft.skills.train', self.train)

        LOG.debug('Loaded Padatious intent pipeline')

    @property
    def padatious_config(self) -> Dict:
        log_deprecation("self.padatious_config is deprecated, access self.config directly instead", "2.0.0")
        return self.config

    @padatious_config.setter
    def padatious_config(self, val):
        log_deprecation("self.padatious_config is deprecated, access self.config directly instead", "2.0.0")
        self.config = val

    def _match_level(self, utterances, limit, lang=None, message: Optional[Message] = None) -> Optional[
        IntentHandlerMatch]:
        """Match intent and make sure a certain level of confidence is reached.

        Args:
            utterances (list of tuples): Utterances to parse, originals paired
                                         with optional normalized version.
            limit (float): required confidence level.
        """
        LOG.debug(f'Padatious Matching confidence > {limit}')
        lang = standardize_lang_tag(lang or self.lang)

        if lang in self.stemmers:
            stemmer = self.stemmers[lang]
        else:
            stemmer = None
        utterances = normalize_utterances(utterances, lang,
                                          stemmer=stemmer,
                                          keep_order=True,
                                          cast_to_ascii=self.remove_punct)
        padatious_intent = self.calc_intent(utterances, lang, message)
        if padatious_intent is not None and padatious_intent.conf > limit:
            skill_id = padatious_intent.name.split(':')[0]
            return IntentHandlerMatch(
                match_type=padatious_intent.name,
                match_data=padatious_intent.matches,
                skill_id=skill_id,
                utterance=padatious_intent.sent)

    def match_high(self, utterances: List[str], lang: str, message: Message) -> Optional[IntentHandlerMatch]:
        """Intent matcher for high confidence.

        Args:
            utterances (list of tuples): Utterances to parse, originals paired
                                         with optional normalized version.
        """
        return self._match_level(utterances, self.conf_high, lang, message)

    def match_medium(self, utterances: List[str], lang: str, message: Message) -> Optional[IntentHandlerMatch]:
        """Intent matcher for medium confidence.

        Args:
            utterances (list of tuples): Utterances to parse, originals paired
                                         with optional normalized version.
        """
        return self._match_level(utterances, self.conf_med, lang, message)

    def match_low(self, utterances: List[str], lang: str, message: Message) -> Optional[IntentHandlerMatch]:
        """Intent matcher for low confidence.

        Args:
            utterances (list of tuples): Utterances to parse, originals paired
                                         with optional normalized version.
        """
        return self._match_level(utterances, self.conf_low, lang, message)

    def train(self, message=None):
        """Perform padatious training.

        Args:
            message (Message): optional triggering message
        """
        # wait for any already ongoing training
        # padatious doesnt like threads
        if not self.finished_training_event.is_set():
            self.finished_training_event.wait()
        with self.lock:
            if not any(engine.must_train for engine in self.containers.values()):
                # LOG.debug(f"Nothing new to train for padatious")
                # inform the rest of the system to not wait for training finish
                self.bus.emit(Message('mycroft.skills.trained'))
                self.finished_training_event.set()
                return
            self.finished_training_event.clear()
            # TODO - run this in subprocess?, sometimes fann2 segfaults and kills ovos-core...
            for lang in self.containers:
                if self.containers[lang].must_train:
                    #LOG.debug(f"Training padatious for lang '{lang}'")
                    self.containers[lang].train()

            # inform the rest of the system to stop waiting for training finish
            self.bus.emit(Message('mycroft.skills.trained'))
            self.finished_training_event.set()

        if not self.first_train.is_set():
            self.first_train.set()

    @deprecated("'wait_and_train' has been deprecated, use 'train' directly", "2.0.0")
    def wait_and_train(self):
        """Wait for minimum time between training and start training."""
        self.train()

    def __detach_intent(self, intent_name):
        """ Remove an intent if it has been registered.

        Args:
            intent_name (str): intent identifier
        """
        if intent_name in self.registered_intents:
            self.registered_intents.remove(intent_name)
            for lang in self.containers:
                for skill_id, intents in self._skill2intent.items():
                    if intent_name in intents:
                        try:
                            if isinstance(self.containers[lang], DomainIntentContainer):
                                self.containers[lang].remove_domain_intent(skill_id, intent_name)
                            else:
                                self.containers[lang].remove_intent(intent_name)
                        except Exception as e:
                            LOG.error(f"Failed to remove intent {intent_name} for skill {skill_id}: {str(e)}")

    def handle_detach_intent(self, message):
        """Messagebus handler for detaching padatious intent.

        Args:
            message (Message): message triggering action
        """
        self.__detach_intent(message.data.get('intent_name'))

    def handle_detach_skill(self, message):
        """Messagebus handler for detaching all intents for skill.

        Args:
            message (Message): message triggering action
        """
        skill_id = message.data.get("skill_id") or message.context.get("skill_id")
        if not skill_id:
            LOG.warning("Skill ID is missing. Detaching all anonymous intents")
            skill_id = "anonymous_skill"
        for i in self._skill2intent[skill_id]:
            self.__detach_intent(i)

    def _unpack_object(self, message):
        """convert message to training data"""
        skill_id = message.data.get("skill_id") or message.context.get("skill_id")
        if not skill_id:
            LOG.warning("Skill ID is missing. Registering under 'anonymous_skill'")
            skill_id = "anonymous_skill"
        file_name = message.data.get('file_name')
        samples = message.data.get("samples")
        name = message.data['name']
        lang = message.data.get('lang', self.lang)
        lang = standardize_lang_tag(lang)
        blacklisted_words = message.data.get('blacklisted_words', [])
        if (not file_name or not isfile(file_name)) and not samples:
            LOG.error('Could not find file ' + file_name)
            return

        if not samples and isfile(file_name):
            with open(file_name) as f:
                samples = [line.strip() for line in f.readlines()]

        samples = deduplicate_list(flatten_list([expand_template(s) for s in samples]))
        if lang in self.stemmers:
            stemmer = self.stemmers[lang]
        else:
            stemmer = None
        samples = normalize_utterances(samples, lang,
                                       stemmer=stemmer,
                                       keep_order=False,
                                       cast_to_ascii=self.remove_punct)
        return lang, skill_id, name, samples, blacklisted_words

    def register_intent(self, message):
        """Messagebus handler for registering intents.

        Args:
            message (Message): message triggering action
        """
        skill_id = message.data.get("skill_id") or message.context.get("skill_id")
        if not skill_id:
            LOG.warning("Skill ID is missing. Registering under 'anonymous_skill'")
            skill_id = message.data["skill_id"] = "anonymous_skill"

        self._skill2intent[skill_id].append(message.data['name'])

        lang = message.data.get('lang', self.lang)
        lang = standardize_lang_tag(lang)
        if lang in self.containers:
            self.registered_intents.append(message.data['name'])
            LOG.debug('Registering Padatious intent: ' + message.data['name'])
            lang, skill_id, name, samples, blacklisted_words = self._unpack_object(message)
            if self.engine_class == DomainIntentContainer:
                self.containers[lang].add_domain_intent(skill_id, name, samples, blacklisted_words)
            else:
                self.containers[lang].add_intent(name, samples, blacklisted_words)

        if self.config.get("instant_train", False) or self.first_train.is_set():
            self.train(message)

    def register_entity(self, message):
        """Messagebus handler for registering entities.

        Args:
            message (Message): message triggering action
        """
        lang = message.data.get('lang', self.lang)
        lang = standardize_lang_tag(lang)
        if lang in self.containers:
            self.registered_entities.append(message.data)
            lang, skill_id, name, samples, _ = self._unpack_object(message)
            LOG.debug('Registering Padatious entity: ' + message.data['name'])
            if self.engine_class == DomainIntentContainer:
                self.containers[lang].add_domain_entity(skill_id, name, samples)
            else:
                self.containers[lang].add_entity(name, samples)

    def calc_intent(self, utterances: Union[str, List[str]], lang: Optional[str] = None,
                    message: Optional[Message] = None) -> Optional[PadatiousIntent]:
        """
        Get the best intent match for the given list of utterances. Utilizes a
        thread pool for overall faster execution. Note that this method is NOT
        compatible with Padatious, but is compatible with Padacioso.
        @param utterances: list of string utterances to get an intent for
        @param lang: language of utterances
        @return:
        """
        if isinstance(utterances, str):
            utterances = [utterances]  # backwards compat when arg was a single string
        utterances = [u for u in utterances if len(u.split()) < self.max_words]
        if not utterances:
            LOG.error(f"utterance exceeds max size of {self.max_words} words, skipping padatious match")
            return None

        lang = lang or self.lang

        lang = self._get_closest_lang(lang)
        if lang is None:  # no intents registered for this lang
            return None

        sess = SessionManager.get(message)

        intent_container = self.containers.get(lang)
        intents = [_calc_padatious_intent(utt, intent_container, sess)
                   for utt in utterances]
        intents = [i for i in intents if i is not None]
        # select best
        if intents:
            return max(intents, key=lambda k: k.conf)

    def _get_closest_lang(self, lang: str) -> Optional[str]:
        if self.containers:
            lang = standardize_lang_tag(lang)
            closest, score = closest_match(lang, list(self.containers.keys()))
            # https://langcodes-hickford.readthedocs.io/en/sphinx/index.html#distance-values
            # 0 -> These codes represent the same language, possibly after filling in values and normalizing.
            # 1- 3 -> These codes indicate a minor regional difference.
            # 4 - 10 -> These codes indicate a significant but unproblematic regional difference.
            if score < 10:
                return closest
        return None

    def shutdown(self):
        self.bus.remove('padatious:register_intent', self.register_intent)
        self.bus.remove('padatious:register_entity', self.register_entity)
        self.bus.remove('intent.service.padatious.get', self.handle_get_padatious)
        self.bus.remove('intent.service.padatious.manifest.get', self.handle_padatious_manifest)
        self.bus.remove('intent.service.padatious.entities.manifest.get', self.handle_entity_manifest)
        self.bus.remove('detach_intent', self.handle_detach_intent)
        self.bus.remove('detach_skill', self.handle_detach_skill)

    def handle_get_padatious(self, message):
        """messagebus handler for perfoming padatious parsing.

        Args:
            message (Message): message triggering the method
        """
        utterance = message.data["utterance"]
        lang = message.data.get("lang", self.lang)
        intent = self.calc_intent(utterance, lang=lang)
        if intent:
            intent = intent.__dict__
        self.bus.emit(message.reply("intent.service.padatious.reply",
                                    {"intent": intent}))

    def handle_padatious_manifest(self, message):
        """Messagebus handler returning the registered padatious intents.

        Args:
            message (Message): message triggering the method
        """
        self.bus.emit(message.reply(
            "intent.service.padatious.manifest",
            {"intents": self.registered_intents}))

    def handle_entity_manifest(self, message):
        """Messagebus handler returning the registered padatious entities.

        Args:
            message (Message): message triggering the method
        """
        self.bus.emit(message.reply(
            "intent.service.padatious.entities.manifest",
            {"entities": self.registered_entities}))


@lru_cache(maxsize=3)  # repeat calls under different conf levels wont re-run code
def _calc_padatious_intent(utt: str,
                           intent_container: Union[IntentContainer, DomainIntentContainer],
                           sess: Session) -> Optional[PadatiousIntent]:
    """
    Try to match an utterance to an intent in an intent_container
    @param utt: str - text to match intent against

    @return: matched PadatiousIntent
    """
    try:
        matches = [m for m in intent_container.calc_intents(utt)
                   if m.name not in sess.blacklisted_intents
                   and m.name.split(":")[0] not in sess.blacklisted_skills]
        if len(matches) == 0:
            return None
        best_match = max(matches, key=lambda x: x.conf)
        best_matches = (
            match for match in matches if match.conf == best_match.conf)
        intent = min(best_matches, key=lambda x: sum(map(len, x.matches.values())))
        intent.sent = utt
        return intent
    except Exception as e:
        LOG.error(e)



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from ovos_padatious.entity_edge import EntityEdge
from ovos_padatious.match_data import MatchData


class PosIntent(object):
    """
    Positional intent
    Used to extract entities

    Args:
        token (str): token to attach to (something like {word})
    """

    def __init__(self, token, intent_name=''):
        self.token = token
        self.edges = [EntityEdge(-1, token, intent_name), EntityEdge(+1, token, intent_name)]

    def match(self, orig_data, entity=None):
        l_matches = [(self.edges[0].match(orig_data.sent, pos), pos)
                     for pos in range(len(orig_data.sent))]
        r_matches = [(self.edges[1].match(orig_data.sent, pos), pos)
                     for pos in range(len(orig_data.sent))]

        def is_valid(l_pos, r_pos):
            if r_pos < l_pos:
                return False
            for p in range(l_pos, r_pos + 1):
                if orig_data.sent[p].startswith('{'):
                    return False
            return True

        possible_matches = []
        for l_conf, l_pos in l_matches:
            if l_conf < 0.2:
                continue
            for r_conf, r_pos in r_matches:
                if r_conf < 0.2:
                    continue
                if not is_valid(l_pos, r_pos):
                    continue

                extracted = orig_data.sent[l_pos:r_pos + 1]

                pos_conf = (l_conf - 0.5 + r_conf - 0.5) / 2 + 0.5
                ent_conf = (entity.match(extracted) if entity else 1)

                new_sent = orig_data.sent[:l_pos] + [self.token] + orig_data.sent[r_pos + 1:]
                new_matches = orig_data.matches.copy()
                new_matches[self.token] = extracted

                extra_conf = math.sqrt(pos_conf * ent_conf) - 0.5
                data = MatchData(orig_data.name, new_sent, new_matches,
                                 orig_data.conf + extra_conf)
                possible_matches.append(data)
        return possible_matches

    def save(self, prefix):
        prefix += '.' + self.token
        for i in self.edges:
            i.save(prefix)

    @classmethod
    def from_file(cls, prefix, token):
        prefix += '.' + token
        self = cls(token)
        for i in self.edges:
            i.load(prefix)
        return self

    def train(self, train_data):
        for i in self.edges:
            i.train(train_data)



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os.path import join

from ovos_padatious.simple_intent import SimpleIntent
from ovos_padatious.trainable import Trainable


class Entity(SimpleIntent, Trainable):
    def __init__(self, name, *args, **kwargs):
        SimpleIntent.__init__(self, name)
        Trainable.__init__(self, name, *args, **kwargs)

    @staticmethod
    def verify_name(token):
        if token[0] in '{}' or token[-1] in '{}':
            raise ValueError('token must not be surrounded in braces (ie. {word} should be word)')

    @staticmethod
    def wrap_name(name):
        """Wraps SkillName:entity into SkillName:{entity}"""
        if ':' in name:
            parts = name.split(':')
            intent_name, ent_name = parts[0], parts[1:]
            return intent_name + ':{' + ':'.join(ent_name) + '}'
        else:
            return '{' + name + '}'

    def save(self, folder):
        prefix = join(folder, self.name)
        SimpleIntent.save(self, prefix)
        self.save_hash(prefix)

    @classmethod
    def from_file(cls, name, folder):
        self = super(Entity, cls).from_file(name, join(folder, name))
        self.load_hash(join(folder, name))
        return self



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
from concurrent.futures import ThreadPoolExecutor
from typing import List
from ovos_padatious.intent import Intent
from ovos_padatious.match_data import MatchData
from ovos_padatious.training_manager import TrainingManager
from ovos_padatious.util import tokenize
from ovos_utils.log import LOG


class IntentManager(TrainingManager):
    """
    Manages intents and performs matching using the OVOS Padatious framework.

    Args:
        cache (str): Path to the cache directory for storing trained models.
    """
    def __init__(self, cache: str, debug: bool = False):
        super().__init__(Intent, cache)
        self.debug = debug

    @property
    def intent_names(self):
        return [i.name for i in self.objects + self.objects_to_train]

    def calc_intents(self, query: str, entity_manager) -> List[MatchData]:
        """
        Calculate matches for the given query against all registered intents.

        Args:
            query (str): The input query to match.
            entity_manager: The entity manager for resolving entities in the query.

        Returns:
            List[MatchData]: A list of matches sorted by confidence.
        """
        sent = tokenize(query)

        def match_intent(intent):
            start_time = time.monotonic()
            try:
                match = intent.match(sent, entity_manager)
                match.detokenize()
                if self.debug:
                    LOG.debug(f"Inference for intent '{intent.name}' took {time.monotonic() - start_time} seconds")
                return match
            except Exception as e:
                LOG.error(f"Error processing intent '{intent.name}': {e}")
                return None

        # Parallelize matching
        with ThreadPoolExecutor() as executor:
            matches = list(executor.map(match_intent, self.objects))

        # Filter out None results from failed matches
        matches = [match for match in matches if match]

        return matches



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from ovos_padatious.match_data import MatchData


class TestMatchData(unittest.TestCase):
    def setUp(self):
        self.match = MatchData('name', ['one', 'two'], {'{word}': ['value', 'tokens']}, 0.5)
        self.sentence = ["it", "'", "s", "a", "new", "sentence"]
        self.sentence2 = ["the", "parents", "'", "house"]

    def test_detokenize(self):
        self.match.detokenize()
        assert self.match.sent == 'one two'

        correct_match = MatchData('name', 'one two', {'word': 'value tokens'}, 0.5)
        assert self.match.__dict__ == correct_match.__dict__

    def test_handle_apostrophes(self):
        joined_sentence = self.match.handle_apostrophes(self.sentence)
        joined_sentence2 = self.match.handle_apostrophes(self.sentence2)
        assert joined_sentence == "it's a new sentence"
        assert joined_sentence2 == "the parents' house"



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MatchData(object):
    """
    A set of data describing how a query fits into an intent

    Attributes:
        name (str): Name of matched intent
        sent (str): The query after entity extraction
        conf (float): Confidence (from 0.0 to 1.0)
        matches (dict of str -> str): Key is the name of the entity and
            value is the extracted part of the sentence
    """
    def __init__(self, name, sent, matches=None, conf=0.0):
        self.name = name
        self.sent = sent
        self.matches = matches or {}
        self.conf = conf

    def __getitem__(self, item):
        return self.matches.__getitem__(item)

    def __contains__(self, item):
        return self.matches.__contains__(item)

    def get(self, key, default=None):
        return self.matches.get(key, default)

    def __repr__(self):
        return repr(self.__dict__)

    @staticmethod
    def handle_apostrophes(old_sentence):
        """
        Attempts to handle utterances with apostrophes in them
        """
        new_sentence = ''
        apostrophe_present = False
        for word in old_sentence:
            if word == "'":
                apostrophe_present = True
                new_sentence += word
            else:
                # If the apostrophe is present we don't want to add
                # a whitespace after the apostrophe
                if apostrophe_present:
                    # If the word after the apostrophe is longer than a character long assume that
                    # the previous word is an "s" + apostrophe instead of "word + apostrophe
                    if len(word) > 1:
                        new_sentence += " " + word
                    else:
                        new_sentence += word
                        apostrophe_present = False
                else:
                    if len(new_sentence) > 0:
                        new_sentence += " " + word
                    else:
                        new_sentence = word
        return new_sentence
    # Converts parameters from lists of tokens to one combined string
    def detokenize(self):
        self.sent = self.handle_apostrophes(self.sent)

        new_matches = {}
        for token, sent in self.matches.items():
            new_token = token.replace('{', '').replace('}', '')
            new_matches[new_token] = self.handle_apostrophes(sent)
        self.matches = new_matches



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from os import mkdir
from os.path import isdir
from shutil import rmtree

from ovos_padatious.intent import Intent
from ovos_padatious.train_data import TrainData


class TestIntent(unittest.TestCase):
    def setUp(self):
        self.data = TrainData()
        self.data.add_lines('hi', ['hello', 'hi', 'hi there'])
        self.data.add_lines('bye', ['goodbye', 'bye', 'bye {person}', 'see you later'])
        self.i_hi = Intent('hi')
        self.i_bye = Intent('bye')
        self.i_hi.train(self.data)
        self.i_bye.train(self.data)

    def test_match(self):
        assert self.i_hi.match(['hi']).conf > self.i_hi.match(['bye']).conf
        assert self.i_hi.match(['hi']).conf > self.i_bye.match(['hi']).conf
        assert self.i_bye.match(['bye']).conf > self.i_bye.match(['hi']).conf
        assert self.i_bye.match(['bye']).conf > self.i_hi.match(['bye']).conf

        all = self.i_bye.match(['see', 'you', 'later']).conf
        assert all > self.i_hi.match(['see']).conf
        assert all > self.i_hi.match(['you']).conf
        assert all > self.i_hi.match(['later']).conf

        matches = self.i_bye.match(['bye', 'john']).matches
        assert len(matches) == 1
        assert '{person}' in matches
        assert matches['{person}'] == ['john']

    def test_save_load(self):
        if not isdir('temp'):
            mkdir('temp')
        self.i_hi.save('temp')
        self.i_bye.save('temp')

        self.i_hi = Intent.from_file('hi', 'temp')
        self.i_bye = Intent.from_file('bye', 'temp')

        self.test_match()

    def teardown(self):
        if isdir('temp'):
            rmtree('temp')



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from os.path import isdir
from shutil import rmtree

from ovos_padatious.id_manager import IdManager


class TestIdManager:
    def test_add(self):
        ids = IdManager()
        assert 'a' not in ids
        ids.add_token('a')
        assert 'a' in ids
        ids.add_token('a')
        assert len(ids) == 1
        ids.add_sent(['b', 'c'])
        assert len(ids) == 3
        for i in ['b', 'c']:
            assert i in ids

    def test_vector(self):
        ids = IdManager()
        assert len(ids.vector()) == 0
        ids.add_token('a')
        assert len(ids.vector()) == 1
        ids.add_token('b')
        vec = ids.vector()
        ids.assign(vec, 'b', 0.5)
        assert vec == [0, 0.5]

    def test_assign(self):
        ids = IdManager(ids={'test': 0, 'word': 1})
        vec = ids.vector()
        ids.assign(vec, 'test', 0.7)
        ids.assign(vec, 'word', 0.2)
        assert vec == [0.7, 0.2]

    def test_save_load(self):
        ids1 = IdManager()
        ids1.add_token('hi')
        ids1.add_token('hello')

        if not isdir('temp'):
            os.mkdir('temp')
        ids1.save('temp/temp')

        ids2 = IdManager()
        ids2.load('temp/temp')

        vec1 = ids1.vector()
        vec2 = ids2.vector()
        ids1.assign(vec1, 'hello', 3)
        ids2.assign(vec2, 'hello', 3)

        assert vec1 == vec2

    def teardown(self):
        if isdir('temp'):
            rmtree('temp')



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ovos_padatious.entity import Entity
from ovos_padatious.training_manager import TrainingManager


class EntityManager(TrainingManager):
    def __init__(self, cache):
        super(EntityManager, self).__init__(Entity, cache)
        self.entity_dict = {}

    def calc_ent_dict(self):
        for i in self.objects:
            self.entity_dict[i.name] = i

    def find(self, intent_name, token):
        local_name, global_name = '', token
        if ':' in intent_name:
            local_name = intent_name.split(':')[0] + ':' + token
        return self.entity_dict.get(local_name, self.entity_dict.get(global_name))

    def remove(self, name):
        name = '{' + name + '}'
        if name in self.entity_dict:
            del self.entity_dict[name]
        super(EntityManager, self).remove(name)



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import os
from functools import wraps
from typing import List, Dict, Any, Optional

from ovos_config.meta import get_xdg_base
from ovos_utils.log import LOG
from ovos_utils.xdg_utils import xdg_data_home

from ovos_padatious import padaos
from ovos_padatious.entity import Entity
from ovos_padatious.entity_manager import EntityManager
from ovos_padatious.intent_manager import IntentManager
from ovos_padatious.match_data import MatchData
from ovos_padatious.util import tokenize
import collections

def _save_args(func):
    """
    Decorator that saves the arguments passed to the function in the serialized_args attribute of the class.

    Args:
        func (function): The function to be decorated.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        func(*args, **kwargs)
        bound_args = inspect.signature(func).bind(*args, **kwargs)
        bound_args.apply_defaults()
        kwargs = bound_args.arguments
        kwargs['__name__'] = func.__name__
        kwargs.pop('self').serialized_args.append(kwargs)

    return wrapper


class IntentContainer:
    """
    Creates an IntentContainer object used to load and match intents

    Args:
        cache_dir (str): Directory for caching the neural network models and intent/entity files.
    """

    def __init__(self, cache_dir: Optional[str] = None, disable_padaos: bool = False) -> None:
        cache_dir = cache_dir or f"{xdg_data_home()}/{get_xdg_base()}/intent_cache"
        os.makedirs(cache_dir, exist_ok=True)
        self.cache_dir: str = cache_dir
        self.must_train: bool = False
        self.intents: IntentManager = IntentManager(cache_dir)
        self.entities: EntityManager = EntityManager(cache_dir)
        self.disable_padaos = disable_padaos
        if self.disable_padaos:
            self.padaos = None
        else:
            self.padaos: padaos.IntentContainer = padaos.IntentContainer()
        self.train_thread: Optional[Any] = None  # deprecated
        self.serialized_args: List[Dict[str, Any]] = []  # Serialized calls for training intents/entities
        self.blacklisted_words: Dict[str, List[str]] = collections.defaultdict(list)

    @property
    def intent_names(self):
        return self.intents.intent_names

    def clear(self) -> None:
        """
        Clears the current intent and entity managers and resets the container.
        """
        os.makedirs(self.cache_dir, exist_ok=True)
        self.must_train = False
        self.intents = IntentManager(self.cache_dir)
        self.entities = EntityManager(self.cache_dir)
        if self.disable_padaos:
            self.padaos = None
        else:
            self.padaos: padaos.IntentContainer = padaos.IntentContainer()
        self.serialized_args = []

    def instantiate_from_disk(self) -> None:
        """
        Instantiates the necessary (internal) data structures when loading persisted model from disk.
        This is done via injecting entities and intents back from cached file versions.
        """
        entity_traindata: Dict[str, List[str]] = {}
        intent_traindata: Dict[str, List[str]] = {}

        # workaround: load training data for both entities and intents since
        # padaos regex needs it for (re)compilation until TODO is cleared
        for f in os.listdir(self.cache_dir):
            if f.endswith('.entity'):
                entity_name = f[0:f.find('.entity')]
                with open(os.path.join(self.cache_dir, f), 'r') as d:
                    entity_traindata[entity_name] = [line.strip()
                                                     for line in d]

            elif f.endswith('.intent'):
                intent_name = f[0:f.find('.intent')]
                with open(os.path.join(self.cache_dir, f), 'r') as d:
                    intent_traindata[intent_name] = [line.strip()
                                                     for line in d]

        # TODO: padaos.compile (regex compilation) is redone when loading: find
        # a way to persist regex, as well!
        for f in os.listdir(self.cache_dir):
            if f.startswith('{') and f.endswith('}.hash'):
                entity_name = f[1:f.find('}.hash')]
                if entity_name in entity_traindata:
                    self.add_entity(
                        name=entity_name,
                        lines=entity_traindata[entity_name],
                        reload_cache=False,
                        must_train=False
                    )
            elif not f.startswith('{') and f.endswith('.hash'):
                intent_name = f[0:f.find('.hash')]
                if intent_name in intent_traindata:
                    self.add_intent(
                        name=intent_name,
                        lines=intent_traindata[intent_name],
                        reload_cache=False,
                        must_train=False
                    )

    @_save_args
    def add_intent(self, name: str, lines: List[str], reload_cache: bool = False, must_train: bool = True,
                   blacklisted_words: Optional[List[str]] = None) -> None:
        """
        Creates a new intent, optionally checking the cache first

        Args:
            name (str): Name of the intent.
            lines (List[str]): Sentences that will activate the intent.
            reload_cache (bool): Whether to ignore cached intent.
            must_train (bool): Whether the model needs training after adding the intent.
        """
        self.blacklisted_words[name] += blacklisted_words or []
        self.intents.add(name, lines, reload_cache, must_train)
        if self.padaos is not None:
            self.padaos.add_intent(name, lines)
        self.must_train = must_train

    @_save_args
    def add_entity(self, name: str, lines: List[str], reload_cache: bool = False, must_train: bool = True) -> None:
        """
        Adds an entity that matches the given lines.

        Example:
            self.add_intent('weather', ['will it rain on {weekday}?'])
            self.add_entity('weekday', ['monday', 'tuesday', 'wednesday'])  # ...

        Args:
            name (str): Name of the entity.
            lines (List[str]): Example extracted entities.
            reload_cache (bool): Whether to refresh the cache.
            must_train (bool): Whether the model needs training after adding the entity.
        """
        Entity.verify_name(name)
        self.entities.add(
            Entity.wrap_name(name),
            lines,
            reload_cache,
            must_train)
        if self.padaos is not None:
            self.padaos.add_entity(name, lines)
        self.must_train = must_train

    @_save_args
    def load_entity(self, name: str, file_name: str, reload_cache: bool = False, must_train: bool = True) -> None:
        """
       Loads an entity, optionally checking the cache first

       Args:
           name (str): The associated name of the entity
           file_name (str): The location of the entity file
           reload_cache (bool): Whether to refresh all of cache
            must_train (bool): Whether the model needs training after loading the entity.
        """
        Entity.verify_name(name)
        self.entities.load(Entity.wrap_name(name), file_name, reload_cache)
        if self.padaos is not None:
            with open(file_name) as f:
                self.padaos.add_entity(name, f.read().split('\n'))
        self.must_train = must_train

    @_save_args
    def load_file(self, *args, **kwargs):
        """Legacy. Use load_intent instead"""
        self.load_intent(*args, **kwargs)

    @_save_args
    def load_intent(self, name: str, file_name: str, reload_cache: bool = False, must_train: bool = True) -> None:
        """
        Loads an intent, optionally checking the cache first

        Args:
            name (str): The associated name of the intent
            file_name (str): The location of the intent file
            reload_cache (bool): Whether to refresh all of cache
            must_train (bool): Whether the model needs training after loading the intent.
        """
        self.intents.load(name, file_name, reload_cache)
        if self.padaos is not None:
            with open(file_name) as f:
                self.padaos.add_intent(name, f.read().split('\n'))
        self.must_train = must_train

    @_save_args
    def remove_intent(self, name: str) -> None:
        """
        Removes an intent by its name.

        Args:
            name (str): Name of the intent to remove.
        """
        self.intents.remove(name)
        if self.padaos is not None:
            self.padaos.remove_intent(name)
        self.must_train = True

    @_save_args
    def remove_entity(self, name: str) -> None:
        """
        Removes an entity by its name.

        Args:
            name (str): Name of the entity to remove.
        """
        self.entities.remove(name)
        if self.padaos is not None:
            self.padaos.remove_entity(name)

    def train(self, debug: bool = True, force: bool = False, single_thread: Optional[bool] = None,
              timeout: Optional[float] = None) -> bool:
        """
        Trains all the loaded intents that need to be updated
        If a cache file exists with the same hash as the intent file,
        the intent will not be trained and just loaded from file

        Args:
            debug (bool): Whether to print a message to stdout each time a new intent is trained
            force (bool): Whether to force training if already finished
            single_thread (bool): DEPRECATED
            timeout (float): DEPRECATED
        Returns:
            bool: True if training succeeded
        """
        if single_thread is not None:
            LOG.warning("'single_thread' argument is deprecated and will be ignored")
        if timeout is not None:
            LOG.warning("'timeout' argument is deprecated and will be ignored")
        if not self.must_train and not force:
            return True

        if self.padaos is not None:
            self.padaos.compile()

        # Train intents and entities
        self.intents.train(debug=debug)
        self.entities.train(debug=debug)

        self.entities.calc_ent_dict()

        self.must_train = False
        return True

    def calc_intents(self, query: str) -> List[MatchData]:
        """
        Tests all the intents against the query and returns
        data on how well each one matched against the query

        Args:
            query (str): Input sentence to test against intents.

        Returns:
            List[MatchData]: A list of all intent matches with confidence scores.
        """
        if self.must_train:
            self.train()
        # post-processing: discard any matches that contain blacklisted words
        intents = {i.name: i
                   for i in self.intents.calc_intents(query, self.entities)
                   if not any(k in query for k in self.blacklisted_words[i.name])}
        sent = tokenize(query)

        if self.padaos is not None:
            for perfect_match in self.padaos.calc_intents(query):
                name = perfect_match['name']
                intents[name] = MatchData(name, sent, matches=perfect_match['entities'], conf=1.0)
        return list(intents.values())

    def calc_intent(self, query: str) -> MatchData:
        """
        Returns the best intent match for the given query.

        Args:
            query (str): Input sentence to test against intents.

        Returns:
            MatchData: The best matching intent.
        """
        matches = self.calc_intents(query)
        if not matches:
            return MatchData('', '')
        best_match = max(matches, key=lambda x: x.conf)
        best_matches = [match for match in matches if match.conf == best_match.conf]
        return min(best_matches, key=lambda x: sum(map(len, x.matches.values())))

    def get_training_args(self) -> List[Dict[str, Any]]:
        """
        Returns all serialized arguments used for training intents and entities.

        Returns:
            List[Dict[str, Any]]: List of serialized arguments for training.
        """
        return self.serialized_args

    def apply_training_args(self, data):
        for params in data:
            func_name = params.pop('__name__')
            getattr(self, func_name)(**params)



# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from unittest import TestCase, mock

from ovos_bus_client.message import Message

from ovos_padatious.opm import PadatiousPipeline


def create_intent_msg(keyword, value):
    """Create a message for registering a padatious intent."""
    return Message('padatious:register_intent',
                   {'name': f"skill:{keyword}", 'samples': [value]},
                   {"skill_id": "skill"})


def create_entity_msg(keyword, value):
    """Create a message for registering a padatious entity."""
    return Message('padatious:register_entity',
                   {'name': f"skill:{keyword}", 'samples': [value]},
                   {"skill_id": "skill"})


def get_last_message(bus):
    """Get last sent message on mock bus."""
    last = bus.emit.call_args
    return last[0][0]


class TestIntentServiceApi(TestCase):
    def setUp(self):
        self.intent_service = PadatiousPipeline(mock.Mock())
        self.setup_simple_padatious_intent()

    def setup_simple_padatious_intent(self,
                                      msg=create_intent_msg('testIntent', 'test'),
                                      msg2=create_entity_msg('testEntity', 'enty')):
        self.intent_service.register_intent(msg)
        self.intent_service.register_entity(msg2)

    def test_get_padatious_intent(self):
        # Check that the intent is returned
        msg = Message('intent.service.padatious.get',
                      data={'utterance': 'test'})
        self.intent_service.handle_get_padatious(msg)

        reply = get_last_message(self.intent_service.bus)

        self.assertEqual(reply.data['intent']['name'],
                         'skill:testIntent')

    def test_get_padatious_intent_no_match(self):
        """Check that if the intent doesn't match at all None is returned."""
        # Check that no intent is matched
        msg = Message('intent.service.padatious.get',
                      data={'utterance': 'five'})
        self.intent_service.handle_get_padatious(msg)
        reply = get_last_message(self.intent_service.bus)
        self.assertLess(reply.data["intent"]['conf'], 0.4)

    def test_get_padatious_intent_manifest(self):
        """Make sure the manifest returns a list of Intent Parser objects."""
        msg = Message('intent.service.padatious.manifest.get')
        self.intent_service.handle_padatious_manifest(msg)
        reply = get_last_message(self.intent_service.bus)
        self.assertEqual(reply.data['intents'][0], 'skill:testIntent')

    def test_get_padatious_vocab_manifest(self):
        msg = Message('intent.service.padatious.entities.manifest.get')
        self.intent_service.handle_entity_manifest(msg)
        reply = get_last_message(self.intent_service.bus)
        value = reply.data["entities"][0]['name']
        keyword = reply.data["entities"][0]['samples'][0]
        self.assertEqual(value, 'skill:testEntity')
        self.assertEqual(keyword, 'enty')

    def test_get_no_match_after_detach(self):
        """Check that a removed intent doesn't match."""
        # Check that no intent is matched
        msg = Message('detach_intent',
                      data={'intent_name': 'skill:testIntent'})
        self.intent_service.handle_detach_intent(msg)
        msg = Message('intent.service.padatious.get', data={'utterance': 'test'})
        self.intent_service.handle_get_padatious(msg)
        reply = get_last_message(self.intent_service.bus)
        self.assertEqual(reply.data['intent'], None)

    def test_get_no_match_after_detach_skill(self):
        """Check that a removed skill's intent doesn't match."""
        # Check that no intent is matched
        msg = Message('detach_intent',
                      data={'skill_id': 'skill'})
        self.intent_service.handle_detach_skill(msg)
        msg = Message('intent.service.padatious.get', data={'utterance': 'test'})
        self.intent_service.handle_get_padatious(msg)
        reply = get_last_message(self.intent_service.bus)
        self.assertEqual(reply.data['intent'], None)



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import random
import unittest
from os.path import join
from time import monotonic

import pytest

from ovos_padatious.intent_container import IntentContainer


class TestFromDisk(unittest.TestCase):
    test_lines = ['this is a test\n', 'another test\n']
    other_lines = ['something else\n', 'this is a different thing\n']
    test_lines_with_entities = ['this is a {test}\n', 'another {test}\n']
    other_lines_with_entities = [
        'something {other}\n',
        'this is a {other} thing\n']
    test_entities = ['test\n', 'assessment\n']
    other_entities = ['else\n', 'different\n']

    def setUp(self):
        self.cont = IntentContainer('/tmp/cache2')

    def _add_intent(self):
        self.cont.add_intent('test', self.test_lines)
        self.cont.add_intent('other', self.other_lines)
        self.cont.add_entity('test', self.test_entities)
        self.cont.add_entity('other', self.other_entities)
        self.cont.train()
        self._write_train_data()

    def _write_train_data(self):
        fn1 = join('/tmp/cache2', 'test.intent')
        with open(fn1, 'w') as f:
            f.writelines(self.test_lines_with_entities)

        fn2 = join('/tmp/cache2', 'other.intent')
        with open(fn2, 'w') as f:
            f.writelines(self.other_lines_with_entities)

        fn1 = join('/tmp/cache2', 'test.entity')
        with open(fn1, 'w') as f:
            f.writelines(self.test_entities)

        fn2 = join('/tmp/cache2', 'other.entity')
        with open(fn2, 'w') as f:
            f.writelines(self.other_entities)

    def test_instantiate_from_disk(self):
        # train and cache (i.e. persist)
        self._add_intent()

        # instantiate from disk (load cached files)
        cont = IntentContainer('/tmp/cache2')
        cont.instantiate_from_disk()

        assert len(cont.intents.train_data.sent_lists) == 0
        assert len(cont.intents.objects_to_train) == 0
        assert len(cont.intents.objects) == 2

        result = cont.calc_intent('something different')
        assert result.matches['other'] == 'different'


class TestIntentContainer(unittest.TestCase):
    test_lines = ['this is a test\n', 'another test\n']
    other_lines = ['something else\n', 'this is a different thing\n']
    test_lines_with_entities = ['this is a {test}\n', 'another {test}\n']
    other_lines_with_entities = [
        'something {other}\n',
        'this is a {other} thing\n']
    test_entities = ['test\n', 'assessment\n']
    other_entities = ['else\n', 'different\n']
    blist = ["bad", "404"]

    def setUp(self):
        self.cont = IntentContainer('/tmp/cache')

    def _add_intent(self, blacklist=False):
        self.cont.add_intent('test', self.test_lines, blacklisted_words=self.blist if blacklist else [])
        self.cont.add_intent('other', self.other_lines, blacklisted_words=self.blist if blacklist else [])

    def test_load_intent(self):
        fn1 = join('/tmp', 'test.txt')
        with open(fn1, 'w') as f:
            f.writelines(self.test_lines)

        fn2 = join('/tmp', 'other.txt')
        with open(fn2, 'w') as f:
            f.writelines(self.other_lines)

        self.cont.load_intent('test', fn1)
        self.cont.load_intent('other', fn1)
        assert len(self.cont.intents.train_data.sent_lists) == 2

    def test_train(self):
        def test(a, b):
            self._add_intent()
            self.cont.train(a, b)

        test(False, False)
        test(True, True)

    def _create_large_intent(self, depth):
        if depth == 0:
            return '(a|b|)'
        return '{0} {0}'.format(self._create_large_intent(depth - 1))

    def test_calc_intents(self):
        self._add_intent()
        self.cont.train(False)

        intents = self.cont.calc_intents('this is another test')
        assert (
                       intents[0].conf > intents[1].conf) == (
                       intents[0].name == 'test')
        assert self.cont.calc_intent('this is another test').name == 'test'

        # this one will fail in next test, should pass here
        intents = self.cont.calc_intents('this is a bad test')
        self.assertEqual(intents[0].name, 'test')

    def test_blacklist(self):
        self._add_intent(blacklist=True)
        self.cont.train(False)
        # matched in previous test
        # no match here due to "bad" being in blacklist
        intents = self.cont.calc_intents('this is a bad test')
        self.assertEqual(intents, [])

    def test_empty(self):
        self.cont.train(False)
        self.cont.calc_intent('hello')

    def _test_entities(self, namespace):
        self.cont.add_intent(namespace + 'intent', [
            'test {ent}'
        ])
        self.cont.add_entity(namespace + 'ent', [
            'one'
        ])
        self.cont.train(False)
        data = self.cont.calc_intent('test one')
        high_conf = data.conf
        assert data.conf > 0.5
        assert data['ent'] == 'one'

        data = self.cont.calc_intent('test two')
        assert high_conf > data.conf
        assert 'ent' not in data

    def test_regular_entities(self):
        self._test_entities('')

    def test_namespaced_entities(self):
        self._test_entities('SkillName:')

    def test_remove(self):
        self._add_intent()
        self.cont.train(False)
        assert self.cont.calc_intent('This is a test').conf == 1.0
        self.cont.remove_intent('test')
        assert self.cont.calc_intent('This is a test').conf < 0.5
        self.cont.add_intent('thing', ['A {thing}'])
        self.cont.add_entity('thing', ['thing'])
        self.cont.train(False)
        assert self.cont.calc_intent('A dog').conf < 0.5
        assert self.cont.calc_intent('A thing').conf == 1.0
        self.cont.remove_entity('thing')
        assert self.cont.calc_intent('A dog').conf == 1.0

    def test_overlap(self):
        self.cont.add_intent('song', ['play {song}'])
        self.cont.add_intent('news', ['play the news'])
        self.cont.train(False)
        assert self.cont.calc_intent('play the news').name == 'news'

    def test_overlap_backwards(self):
        self.cont.add_intent('song', ['play {song}'])
        self.cont.add_intent('news', ['play the news'])
        self.cont.train(False)
        assert self.cont.calc_intent('play the news').name == 'news'

    def test_generalize(self):
        self.cont.add_intent('timer', [
            'set a timer for {time} minutes',
            'make a {time} minute timer'
        ])
        self.cont.add_entity('time', [
            '#', '##', '#:##', '##:##'
        ])
        self.cont.train(False)
        intent = self.cont.calc_intent('make a timer for 3 minute')
        assert intent.name == 'timer'
        assert intent.matches == {'time': '3'}



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
import os
from os.path import isfile

from ovos_padatious.train_data import TrainData


class TestTrainData(unittest.TestCase):
    def setUp(self):
        self.data = TrainData()
        with open('/tmp/train', 'w') as f:
            f.writelines(['hi'])

    def test_add_lines(self):
        self.data.add_file('hi', '/tmp/train')
        self.data.add_lines('bye', ['bye'])
        self.data.add_lines('other', ['other'])

        def cmp(a, b):
            return set(' '.join(i) for i in a) == set(' '.join(i) for i in b)

        assert cmp(self.data.my_sents('hi'), [['hi']])
        assert cmp(self.data.other_sents('hi'), [['bye'], ['other']])
        assert cmp(self.data.all_sents(), [['hi'], ['bye'], ['other']])

    def teardown(self):
        if isfile('/tmp/train'):
            os.remove('/tmp/train')



import inspect
import json
from os.path import basename, splitext

from argparse import ArgumentParser

from ovos_padatious import IntentContainer


def train_setup(parser):
    parser.add_argument('intent_cache', help='Folder to write trained intents to')
    parser.add_argument('input_files', nargs='*', help='Input .intent and .entity files')
    parser.add_argument('-d', '--data', help='Serialized training args', type=json.loads)
    parser.add_argument('-s', '--single-thread', help='Run training in a single thread')
    parser.add_argument('-f', '--force', help='Force retraining if already trained')
    parser.add_argument('-a', '--args', help='Extra args (list) for function', type=json.loads)
    parser.add_argument('-k', '--kwargs', help='Extra kwargs (json) for function', type=json.loads)


def train(parser, args):
    if bool(args.input_files) == bool(args.data):
        parser.error('You must specify one of input_files or --data (but not both)')

    cont = IntentContainer(args.intent_cache)
    if args.data:
        cont.apply_training_args(args.data)
    else:
        for fn in args.input_files:
            obj_name, ext = splitext(basename(fn))
            if ext == '.intent':
                cont.load_intent(obj_name, fn)
            elif ext == '.entity':
                cont.load_entity(obj_name, fn)
            else:
                parser.error('Unknown file extension: {}'.format(ext))
    kwargs = inspect.signature(cont.train).bind(*(args.args or [])).arguments
    kwargs.update(args.kwargs or {})
    kwargs.setdefault('debug', True)
    kwargs.setdefault('single_thread', args.single_thread)
    kwargs.setdefault('force', args.force)
    if cont.train(**kwargs):
        return 0
    return 10  # timeout


def main():
    parser = ArgumentParser(description='Tool to interact with ovos_padatious via command line')
    p = parser.add_subparsers(dest='action')
    p.required = True
    train_setup(p.add_parser('train'))

    args = parser.parse_args()
    if args.action == 'train':
        exit(train(parser, args))


if __name__ == '__main__':
    main()



import unittest
from unittest.mock import MagicMock

from ovos_padatious.domain_container import DomainIntentContainer  # Replace 'your_module' with the actual module name

from ovos_padatious.match_data import MatchData


class TestDomainIntentEngine(unittest.TestCase):
    def setUp(self):
        self.engine = DomainIntentContainer()

    def test_register_domain_intent(self):
        self.engine.add_domain_intent("domain1", "intent1", ["sample1", "sample2"])
        self.assertIn("domain1", self.engine.training_data)
        self.assertIn("intent1", self.engine.domains["domain1"].intent_names)

    def test_remove_domain(self):
        self.engine.add_domain_intent("domain1", "intent1", ["sample1", "sample2"])
        self.engine.remove_domain("domain1")
        self.assertNotIn("domain1", self.engine.training_data)
        self.assertNotIn("domain1", self.engine.domains)

    def test_remove_domain_intent(self):
        self.engine.add_domain_intent("domain1", "intent1", ["sample1", "sample2"])
        self.engine.remove_domain_intent("domain1", "intent1")
        self.assertNotIn("intent1", self.engine.domains["domain1"].intent_names)

    def test_calc_domains(self):
        self.engine.train = MagicMock()
        self.engine.domain_engine.calc_intents = MagicMock(
            return_value=[MatchData(name="domain1", sent="query", matches=None, conf=0.9)])
        result = self.engine.calc_domains("query")
        self.engine.train.assert_called_once()
        self.assertEqual(result[0].name, "domain1")

    def test_calc_domain(self):
        self.engine.train = MagicMock()
        self.engine.domain_engine.calc_intent = MagicMock(
            return_value=MatchData(name="domain1", sent="query", matches=None, conf=0.9))
        result = self.engine.calc_domain("query")
        self.engine.train.assert_called_once()
        self.assertEqual(result.name, "domain1")

    def test_calc_intent(self):
        self.engine.train = MagicMock()
        mock_domain_container = MagicMock()
        mock_domain_container.calc_intent.return_value = MatchData(name="intent1", sent="query", matches=None, conf=0.9)
        self.engine.domains["domain1"] = mock_domain_container

        self.engine.domain_engine.calc_intent = MagicMock(
            return_value=MatchData(name="domain1", sent="query", matches=None, conf=0.9))
        result = self.engine.calc_intent("query")
        self.assertEqual(result.name, "intent1")

    def test_calc_intents(self):
        self.engine.train = MagicMock()
        mock_domain_container = MagicMock()
        mock_domain_container.calc_intents.return_value = [
            MatchData(name="intent1", sent="query", matches=None, conf=0.9),
            MatchData(name="intent2", sent="query", matches=None, conf=0.8),
        ]
        self.engine.domains["domain1"] = mock_domain_container

        self.engine.domain_engine.calc_intents = MagicMock(
            return_value=[MatchData(name="domain1", sent="query", matches=None, conf=0.9)])
        result = self.engine.calc_intents("query")
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0].name, "intent1")

    def test_train(self):
        self.engine.training_data["domain1"] = ["sample1", "sample2"]
        self.engine.domain_engine.add_intent = MagicMock()
        self.engine.domain_engine.train = MagicMock()

        mock_domain_container = MagicMock()
        self.engine.domains["domain1"] = mock_domain_container

        self.engine.train()
        self.engine.domain_engine.add_intent.assert_called_with("domain1", ["sample1", "sample2"])
        self.engine.domain_engine.train.assert_called_once()
        mock_domain_container.train.assert_called_once()
        self.assertFalse(self.engine.must_train)


class TestDomainIntentEngineWithLiveData(unittest.TestCase):
    def setUp(self):
        self.engine = DomainIntentContainer()
        # Sample training data
        self.training_data = {
            "IOT": {
                "turn_on_device": ["Turn on the lights", "Switch on the fan", "Activate the air conditioner"],
                "turn_off_device": ["Turn off the lights", "Switch off the heater", "Deactivate the air conditioner"],
            },
            "greetings": {
                "say_hello": ["Hello", "Hi there", "Good morning"],
                "say_goodbye": ["Goodbye", "See you later", "Bye"],
            },
            "General Knowledge": {
                "ask_fact": ["Tell me a fact about space", "What is the capital of France?",
                             "Who invented the telephone?"],
            },
            "Question": {
                "ask_question": ["Why is the sky blue?", "What is quantum mechanics?",
                                 "Can you explain photosynthesis?"],
            },
            "Media Playback": {
                "play_music": ["Play some music", "Start the playlist", "Play a song"],
                "stop_music": ["Stop the music", "Pause playback", "Halt the song"],
            },
        }
        # Register domains and intents
        for domain, intents in self.training_data.items():
            for intent, samples in intents.items():
                self.engine.add_domain_intent(domain, intent, samples)
        self.engine.train()

    def test_live_data_intent_matching(self):
        # Test IOT domain
        query = "Switch on the fan"
        result = self.engine.calc_intent(query, domain="IOT")
        self.assertEqual(result.name, "turn_on_device")
        self.assertGreater(result.conf, 0.8)

        # Test greetings domain
        query = "Hi there"
        result = self.engine.calc_intent(query, domain="greetings")
        self.assertEqual(result.name, "say_hello")
        self.assertGreater(result.conf, 0.8)

        # Test General Knowledge domain
        query = "What is the capital of France?"
        result = self.engine.calc_intent(query, domain="General Knowledge")
        self.assertEqual(result.name, "ask_fact")
        self.assertGreater(result.conf, 0.8)

        # Test Question domain
        query = "Why is the sky blue?"
        result = self.engine.calc_intent(query, domain="Question")
        self.assertEqual(result.name, "ask_question")
        self.assertGreater(result.conf, 0.8)

        # Test Media Playback domain
        query = "Play a song"
        result = self.engine.calc_intent(query, domain="Media Playback")
        self.assertEqual(result.name, "play_music")
        self.assertGreater(result.conf, 0.8)

    def test_live_data_cross_domain_matching(self):
        # Test cross-domain intent matching
        query = "Tell me a fact about space"
        result = self.engine.calc_domain(query)
        self.assertEqual(result.name, "General Knowledge")
        self.assertGreater(result.conf, 0.8)

        # Validate intent from the matched domain
        result = self.engine.calc_intent(query, domain=result.name)
        self.assertEqual(result.name, "ask_fact")
        self.assertGreater(result.conf, 0.8)

    def test_calc_intent_without_domain(self):
        # Test intent calculation without specifying a domain
        query = "Turn on the lights"
        result = self.engine.calc_intent(query)
        self.assertIsNotNone(result.name, "Intent name should not be None")
        self.assertEqual(result.name, "turn_on_device")
        self.assertGreater(result.conf, 0.8)

        query = "Goodbye"
        result = self.engine.calc_intent(query)
        self.assertIsNotNone(result.name, "Intent name should not be None")
        self.assertEqual(result.name, "say_goodbye")
        self.assertGreater(result.conf, 0.8)

        query = "What is quantum mechanics?"
        result = self.engine.calc_intent(query)
        self.assertIsNotNone(result.name, "Intent name should not be None")
        self.assertEqual(result.name, "ask_question")
        self.assertGreater(result.conf, 0.8)


if __name__ == "__main__":
    unittest.main()



# Copyright 2017 Mycroft AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import partial
from os.path import join, isfile, splitext
from typing import List, Type, Union

from ovos_utils.log import LOG

import ovos_padatious
from ovos_padatious.trainable import Trainable
from ovos_padatious.train_data import TrainData
from ovos_padatious.util import lines_hash


def _train_and_save(obj: Trainable, cache: str, data: TrainData, print_updates: bool) -> None:
    """
    Internal function to train objects sequentially and save them.

    Args:
        obj (Trainable): Object to train (Intent or Entity).
        cache (str): Path to the cache directory.
        data (TrainData): Training data.
        print_updates (bool): Whether to print updates during training.
    """
    obj.train(data)
    obj.save(cache)
    if print_updates:
        LOG.debug(f'Saving {obj.name} to cache ({cache})')


class TrainingManager:
    """
    Manages sequential training of either Intents or Entities.

    Args:
        cls (Type[Trainable]): Class to wrap (Intent or Entity).
        cache_dir (str): Path to the cache directory.
    """

    def __init__(self, cls: Type[Trainable], cache_dir: str) -> None:
        """
        Initializes the TrainingManager.

        Args:
            cls (Type[Trainable]): Class to be managed (Intent or Entity).
            cache_dir (str): Path where cache files are stored.
        """
        self.cls = cls
        self.cache = cache_dir
        self.objects: List[Trainable] = []
        self.objects_to_train: List[Trainable] = []
        self.train_data = TrainData()

    def add(self, name: str, lines: List[str], reload_cache: bool = False, must_train: bool = True) -> None:
        """
        Adds a new intent or entity for training or loading from cache.

        Args:
            name (str): Name of the intent or entity.
            lines (List[str]): Lines of training data.
            reload_cache (bool): Whether to force reload of cache if it exists.
            must_train (bool): Whether training is required for the new intent/entity.
        """
        if not must_train:
            LOG.debug(f"Loading {name} from intent cache")
            self.objects.append(self.cls.from_file(name=name, folder=self.cache))
        # general case: load resource (entity or intent) to training queue
        # or if no change occurred to memory data structures
        else:
            hash_fn = join(self.cache, name + '.hash')
            old_hsh = None
            min_ver = splitext(ovos_padatious.__version__)[0]
            new_hsh = lines_hash([min_ver] + lines)

            if isfile(hash_fn):
                with open(hash_fn, 'rb') as g:
                    old_hsh = g.read()
                if old_hsh != new_hsh:
                    LOG.debug(f"{name} training data changed! retraining")
            else:
                LOG.debug(f"First time training '{name}")

            retrain = reload_cache or old_hsh != new_hsh
            if not retrain:
                try:
                    LOG.debug(f"Loading {name} from intent cache")
                    self.objects.append(self.cls.from_file(name=name, folder=self.cache))
                except Exception as e:
                    LOG.error(f"Failed to load intent from cache: {name} - {str(e)}")
                    retrain = True
            if retrain:
                LOG.debug(f"Queuing {name} for training")
                self.objects_to_train.append(self.cls(name=name, hsh=new_hsh))
            self.train_data.add_lines(name, lines)

    def load(self, name: str, file_name: str, reload_cache: bool = False) -> None:
        """
        Loads an entity or intent from a file and adds it for training or caching.

        Args:
            name (str): Name of the intent or entity.
            file_name (str): Path to the file containing the training data.
            reload_cache (bool): Whether to reload the cache for this intent/entity.
        """
        with open(file_name) as f:
            self.add(name, f.read().split('\n'), reload_cache)

    def remove(self, name: str) -> None:
        """
        Removes an intent or entity from the training and cache.

        Args:
            name (str): Name of the intent or entity to remove.
        """
        self.objects = [i for i in self.objects if i.name != name]
        self.objects_to_train = [i for i in self.objects_to_train if i.name != name]
        self.train_data.remove_lines(name)

    def train(self, debug: bool = True, single_thread: Union[None, bool] = None,
              timeout: Union[None, int] = None) -> None:
        """
        Trains all intents and entities sequentially.

        Args:
            debug (bool): Whether to print debug messages.
            single_thread (bool): DEPRECATED
            timeout (float): DEPRECATED
        """
        if single_thread is not None:
            LOG.warning("'single_thread' argument is deprecated and will be ignored")
        if timeout is not None:
            LOG.warning("'timeout' argument is deprecated and will be ignored")

        train = partial(_train_and_save, cache=self.cache, data=self.train_data, print_updates=debug)

        objs = list(self.objects_to_train) # make a copy so its thread safe
        fails = []
        # Train objects sequentially
        for obj in objs:
            try:
                train(obj)
            except Exception as e:
                LOG.error(f"Error training {obj.name}: {e}")
                fails.append(obj)

        # Load saved objects from disk
        for obj in objs:
            try:
                self.objects.append(self.cls.from_file(name=obj.name, folder=self.cache))
            except Exception as e:
                LOG.error(f"Failed to load trained object {obj.name}: {e}")
                fails.append(obj)
        self.objects_to_train = [o for o in self.objects_to_train
                                 if o not in objs or o in fails]



from collections import defaultdict
from typing import Dict, List, Optional
from ovos_utils.log import LOG
from ovos_padatious.intent_container import IntentContainer
from ovos_padatious.match_data import MatchData


class DomainIntentContainer:
    """
    A domain-aware intent recognition engine that organizes intents and entities
    into specific domains, providing flexible and hierarchical intent matching.
    """

    def __init__(self, cache_dir: Optional[str] = None, disable_padaos: bool = False):
        """
        Initialize the DomainIntentEngine.

        Attributes:
            domain_engine (IntentContainer): A top-level intent container for cross-domain calculations.
            domains (Dict[str, IntentContainer]): A mapping of domain names to their respective intent containers.
            training_data (Dict[str, List[str]]): A mapping of domain names to their associated training samples.
        """
        self.cache_dir = cache_dir
        self.disable_padaos = disable_padaos
        self.domain_engine = IntentContainer(cache_dir=cache_dir,
                                             disable_padaos=disable_padaos)
        self.domains: Dict[str, IntentContainer] = {}
        self.training_data: Dict[str, List[str]] = defaultdict(list)
        self.instantiate_from_disk()
        self.must_train = True

    def instantiate_from_disk(self) -> None:
        """
        Instantiates the necessary (internal) data structures when loading persisted model from disk.
        This is done via injecting entities and intents back from cached file versions.
        """
        self.domain_engine.instantiate_from_disk()
        for engine in self.domains.values():
            engine.instantiate_from_disk()

    def remove_domain(self, domain_name: str):
        """
        Remove a domain and its associated intents and training data.

        Args:
            domain_name (str): The name of the domain to remove.
        """
        if domain_name in self.training_data:
            self.training_data.pop(domain_name)
        if domain_name in self.domains:
            self.domains.pop(domain_name)
        if domain_name in self.domain_engine.intent_names:
            self.domain_engine.remove_intent(domain_name)

    def add_domain_intent(self, domain_name: str, intent_name: str, intent_samples: List[str],
                          blacklisted_words: Optional[List[str]] = None):
        """
        Register an intent within a specific domain.

        Args:
            domain_name (str): The name of the domain.
            intent_name (str): The name of the intent to register.
            intent_samples (List[str]): A list of sample sentences for the intent.
        """
        if domain_name not in self.domains:
            self.domains[domain_name] = IntentContainer(cache_dir=self.cache_dir,
                                                        disable_padaos=self.disable_padaos)
            self.domains[domain_name].instantiate_from_disk()

        self.domains[domain_name].add_intent(intent_name, intent_samples,
                                             blacklisted_words=blacklisted_words)
        self.training_data[domain_name] += intent_samples
        self.must_train = True

    def remove_domain_intent(self, domain_name: str, intent_name: str):
        """
        Remove a specific intent from a domain.

        Args:
            domain_name (str): The name of the domain.
            intent_name (str): The name of the intent to remove.
        """
        if domain_name in self.domains:
            self.domains[domain_name].remove_intent(intent_name)

    def add_domain_entity(self, domain_name: str, entity_name: str, entity_samples: List[str]):
        """
        Register an entity within a specific domain.

        Args:
            domain_name (str): The name of the domain.
            entity_name (str): The name of the entity to register.
            entity_samples (List[str]): A list of sample phrases for the entity.
        """
        if domain_name not in self.domains:
            self.domains[domain_name] = IntentContainer(cache_dir=self.cache_dir,
                                                        disable_padaos=self.disable_padaos)
        self.domains[domain_name].add_entity(entity_name, entity_samples)

    def remove_domain_entity(self, domain_name: str, entity_name: str):
        """
        Remove a specific entity from a domain.

        Args:
            domain_name (str): The name of the domain.
            entity_name (str): The name of the entity to remove.
        """
        if domain_name in self.domains:
            self.domains[domain_name].remove_entity(entity_name)

    def calc_domains(self, query: str) -> List[MatchData]:
        """
        Calculate the matching domains for a query.

        Args:
            query (str): The input query.

        Returns:
            List[MatchData]: A list of MatchData objects representing matching domains.
        """
        if self.must_train:
            self.train()

        return self.domain_engine.calc_intents(query)

    def calc_domain(self, query: str) -> MatchData:
        """
        Calculate the best matching domain for a query.

        Args:
            query (str): The input query.

        Returns:
            MatchData: The best matching domain.
        """
        if self.must_train:
            self.train()
        return self.domain_engine.calc_intent(query)

    def calc_intent(self, query: str, domain: Optional[str] = None) -> MatchData:
        """
        Calculate the best matching intent for a query within a specific domain.

        Args:
            query (str): The input query.
            domain (Optional[str]): The domain to limit the search to. Defaults to None.

        Returns:
            MatchData: The best matching intent.
        """
        if self.must_train:
            self.train()
        domain: str = domain or self.domain_engine.calc_intent(query).name
        if domain in self.domains:
            return self.domains[domain].calc_intent(query)
        return MatchData(name=None, sent=query, matches=None, conf=0.0)

    def calc_intents(self, query: str, domain: Optional[str] = None, top_k_domains: int = 2) -> List[MatchData]:
        """
        Calculate matching intents for a query across domains or within a specific domain.

        Args:
            query (str): The input query.
            domain (Optional[str]): The specific domain to search in. If None, searches across top-k domains.
            top_k_domains (int): The number of top domains to consider. Defaults to 2.

        Returns:
            List[MatchData]: A list of MatchData objects representing matching intents, sorted by confidence.
        """
        if self.must_train:
            self.train()
        if domain:
            return self.domains[domain].calc_intents(query)
        matches = []
        domains = self.calc_domains(query)[:top_k_domains]
        for domain in domains:
            if domain.name in self.domains:
                matches += self.domains[domain.name].calc_intents(query)
        return sorted(matches, reverse=True, key=lambda k: k.conf)

    def train(self):
        for domain, samples in dict(self.training_data).items():  # copy for thread safety
            LOG.debug(f"Training domain: {domain}")
            self.domain_engine.add_intent(domain, samples)
        self.domain_engine.train()
        for domain in dict(self.domains): # copy for thread safety
            LOG.debug(f"Training domain sub-intents: {domain}")
            self.domains[domain].train()
        self.must_train = False
